package jirc;

import java.util.regex.*;
import java.io.*;
import java.net.*;
import java.util.*;

/**
 * JIRC is a easy way to implement irc functionality in your programs.
 * Create an implementation, then use JIRC.start() to start the irc connection
 * @author ricky
 */
public abstract class JIRC extends Thread{
    private Pattern QuitMessage = Pattern.compile(":([^!]+)!([^@]+)@([\\S]+) QUIT :(.*)");
    private Pattern PartMessage = Pattern.compile(":([^!]+)!([^@]+)@([\\S]+) PART #(.+)");
    private Pattern JoinMessage = Pattern.compile(":([^!]+)!([^@]+)@([\\S]+) JOIN :#(.+)");
    private Pattern PrivMessage = Pattern.compile(":([^!]+)!([^@]+)@([\\S]+) PRIVMSG (#?)(\\S+)\\s:(.+)");
    private Pattern ModeMessage = Pattern.compile(":([^!]+)!([^@]+)@([\\S]+) MODE #([\\S]+) ([\\S]+) (.+)");
    private Pattern NamesMessage = Pattern.compile(":([\\S]+) (353) [\\S]+ (?:=|@) #([\\S]+) :(.+)");
    private Pattern EndNamesMessage = Pattern.compile(":([\\S]+) 366 [\\S]+ (?:=|@) #([\\S]+) :(.+)");
    private Pattern NickInUseError = Pattern.compile(":([\\S]+) 433 * (.*?) :(.+)");
    private Pattern PingMessage = Pattern.compile("PING :(.+)");
    private Pattern PongMessage = Pattern.compile(":([\\S]+) PONG (.*?) :(.+)");
    private String IRCServerName;
    private String IRCNickName;
    private String IRCRealName;
    private List<Channel> Channels = new LinkedList<Channel>();
    private Socket echoSocket;
    private PrintWriter out;
    private BufferedReader in;
    /**
     * A variable holding System.out
     */
    protected PrintStream sout = System.out;
    /**
     * A variable holding System.in
     */
    protected BufferedReader sin = new BufferedReader(new InputStreamReader(System.in));
    private String[] RSInfo = null;
    private DisconnectListenerThread dlt;
    private boolean pongreply = false;
    
    protected abstract void onConnect();
    protected abstract void onPingged(String server);
    protected abstract void onServerMessage(String message);
    protected abstract void onChannelMessageReceive(String channel, String nick, String message);
    protected abstract void onPrivateMessageReceive(String nick, String message);
    protected abstract void onChannelMemberPart(String channel, String nick);
    protected abstract void onChannelMemberJoin(String channel, String nick);
    protected abstract void onChannelMemberQuit(String nick, String message);
    
    /**
     * 
     * @param IRCServerName Name of the server to connect to (eg efnet.xs4all.nl or 127.0.0.1)
     * @param IRCNickName Nick name to connect to the server with
     * @param IRCRealName "Real" name to show tho the server
     */
    public JIRC(String IRCServerName, String IRCNickName, String IRCRealName) {
        this.IRCServerName = IRCServerName;
        this.IRCNickName = IRCNickName;
        this.IRCRealName = IRCRealName;
    }
    
    public void run() {
	try {
            echoSocket = new Socket(IRCServerName, 6666);
            out = new PrintWriter(echoSocket.getOutputStream(), true);
            in = new BufferedReader(new InputStreamReader(echoSocket.getInputStream()));
            out.println("NICK " + IRCNickName);
            out.println("USER " + IRCNickName + " 0 * :" + IRCRealName);
	    if (RSInfo != null) {
		for (int x = 0; x < firstNull(RSInfo); x++) {
		    joinChannel(RSInfo[x]);
		}
		RSInfo = null;
		dlt.connected = true;
	    } else {
		dlt = new DisconnectListenerThread();
		dlt.start();
		dlt.connected = true;
	    }
            onConnect();
	    while (true) {
		Thread.sleep(10);
                if (RSInfo != null) run();
		try {
		    while (in.ready()) {
			String Input = in.readLine();
			if (Input != null) {
			    Input = Input.replaceAll("\\x03(?:[0-9][0-9]|[0-9])", "");
			    Matcher m;
			    if ((m = PingMessage.matcher(Input)).find()) {
				out.println("PONG " + m.group(1));
                                onPingged(m.group(1));
			    } else if ((m = PongMessage.matcher(Input)).find()) {
				pongreply = true;
			    } else if ((m = PrivMessage.matcher(Input)).find() && m.group(4).equals("#")) {
				String msg = m.group(6);
				String channel = m.group(5);
				String nick = m.group(1);
                                onChannelMessageReceive(channel, nick, msg);
			    } else if ((m = PrivMessage.matcher(Input)).find() && m.group(5).equalsIgnoreCase(IRCNickName)) {
				String msg = m.group(6);
				String nick = m.group(1);
                                onPrivateMessageReceive(nick, msg);
			    } else if ((m = QuitMessage.matcher(Input)).find()) {
				String msg = m.group(4);
				String hostmask = m.group(3);
				String name = m.group(2);
				String nick = m.group(1);
                                //hmhmmhmhmmhmmmmmmmm, recheck for all?...
                                //What was I asking there????? ^^
                                onChannelMemberQuit(nick, msg);
			    } else if ((m = PartMessage.matcher(Input)).find()) {
				String channel = m.group(4);
				String hostmask = m.group(3);
				String name = m.group(2);
				String nick = m.group(1);
				out.println("NAMES #" + channel);
                                onChannelMemberPart(channel, nick);
			    } else if ((m = JoinMessage.matcher(Input)).find()) {
				String channel = m.group(4);
				String hostmask = m.group(3);
				String name = m.group(2);
				String nick = m.group(1);
				out.println("NAMES #" + channel);
                                onChannelMemberJoin(channel, nick);
			    } else if ((m = ModeMessage.matcher(Input)).find()) {
				String channel = m.group(4);
				String hostmask = m.group(3);
				String name = m.group(2);
				String nick = m.group(1);
				String modes = m.group(5);
				String targets = m.group(6);
				out.println("NAMES #" + channel);
				sout.println(nick + " (" + name + "@" + hostmask + ") set " + modes + " on " + targets);
			    } else if ((m = NamesMessage.matcher(Input)).find()) {
				String server = m.group(1);
				String channel = m.group(3);
				String[] names = m.group(4).split(" ");
				int temp = inChannelList(Channels, channel);
				if (temp == -1) {
				    Person[] people = new Person[names.length];
				    for (int x = 0; x < names.length; x++) {
					people[x] = new Person(names[x]);
				    }
				    Channels.add(new Channel(channel, people));
				} else {
				    for (int x = 0; x < Channels.get(temp).getAllMembers().length; x++) {
					Channels.get(temp).getAllMembers()[x].makeInactive();
				    }
				    for (int x = 0; x < names.length; x++) {
					if (Channels.get(temp).isMember(names[x].replace("@", "").replace("+", "")) == -1) Channels.get(temp).addMember(names[x]);
					Channels.get(temp).makeActive(names[x].replace("@", "").replace("+", ""));
				    }
				}
			    } else if ((m = NickInUseError.matcher(Input)).find()) {
				String server = m.group(1);
				String nick = m.group(2);
				String msg = m.group(3);
                                changeNick("_" + getNick() + "_");
                                sout.println("Nick in use, changed to " + getNick());
			    } else {
                                onServerMessage(Input);
			    }
			}
		    }
		} catch (Exception e) {
		    e.printStackTrace();
		}
	    }
        } catch (UnknownHostException e) {
            System.out.println("Don't know about host: " + IRCServerName + ". Trying again in 5 seconds...");
            try {sleep(10000);} catch (Exception ex) {}
            run();
        } catch (IOException e) {
            System.out.println("Couldn't get I/O for the connection to: " + IRCServerName + ". Trying again in 5 seconds...");
            try {sleep(10000);} catch (Exception ex) {}
            run();
        } catch (Exception e) {
	    e.printStackTrace();
	}
    }
    
    /**
     * Sends a message to the specified channel
     * <br><i>Note: channels are passed around sans-pound, so you would pass the channel "test", not the channel "#test"</i>
     * @param channel The channel to send the message to
     * @param msg The message to send
     */
    public void sendMessage(String channel, String msg) {
	out.println("PRIVMSG #" + channel + " :" + msg);
    }
    
    /**
     * Sends raw text to the server.  Useful for sending things not yet implemented
     * @param message the message to send (eg NICK newnick)
     */
    public void sendRawMessage(String message) {
	out.println(message);
    }
    
    /**
     * Sends a message to the specified person
     * @param person The person to send the message to
     * @param msg The message to send
     */
    public void sendPrivateMessage(String person, String msg) {
	out.println("PRIVMSG " + person + " :" + msg);
    }
    
    /**
     * Joins the specified channel, allowing you to send and receive messages from it
     * <br><i>Note: channels are passed around sans-pound, so you would pass the channel "test", not the channel "#test"</i>
     * @param channel The channel to join
     */
    public void joinChannel(String channel) {
	out.println("JOIN #" + channel);
    }
    
    /**
     * Parts the specified channel, meaning you can no longer get and receive messages from it
     * <br><i>Note: channels are passed around sans-pound, so you would pass the channel "test", not the channel "#test"</i>
     * @param channel The channel to join
     */
    public void partChannel(String channel) {
        out.println("PART #" + channel);
        int temp = inChannelList(Channels, channel);
        if (temp > -1) {
            Channels.remove(temp);
        } else {
            sout.println("Apparently, \"" + channel + "\" was not in Channels[]");
        }
    }
    
    /**
     * Changes the current nickname of the connection
     * @param nick What to change the nickname to
     */
    public void changeNick(String nick) {
        out.println("NICK " + nick);
        IRCNickName = nick;
    }
    
    /**
     * Gets the current nickname of the connection
     * @return The current nickname
     */
    public String getNick() {
        return IRCNickName;
    }
    
    /**
     * Sets operator status on the given channel member, assuming you have operator status
     * @param channel The channel to change the person is in
     * @param name The name of the person to give operator status to
     */
    public void opChannelMember(String channel, String name) {
        out.println("MODE #" + channel + " +o " + name);
    }
    
    /**
     * Removes operator status on the given channel member, assuming you have operator status
     * @param channel The channel to change the person is in
     * @param name The name of the person to remove operator status from
     */
    public void deOpChannelMember(String channel, String name) {
        out.println("MODE #" + channel + " -o " + name);
    }
    
    private void reconnect() {
	sout.println("Reconnecting...");
	try {
	    RSInfo = new String[Channels.size()];
	    for (int x = 0; x < Channels.size(); x++) {
		RSInfo[x] = Channels.get(x).Name;
	    }
	    out.close();
	    in.close();
	} catch (Exception e) {
	    e.printStackTrace(sout);
	    sout.println("I guess not...");
	}
    }
    
    /**
     * Gets an array version of the List of Channel objects used by JIRC
     * @return The array of Channels
     */
    public Channel[] getChannelArray() {
        return (Channel[])Channels.toArray();
    }
    
    /**
     * Gets the List of Channel objects used by JIRC
     * @return The List of Channels
     */
    public List<Channel> getChannelList() {
        return Channels;
    }
    
    /**
     * 
     * @param list
     * @param match
     * @return
     */
    public static int inChannelList(List<Channel> list, String match) {
	if (match.length() == 0) return -1;
	    for (int x = 0; x < list.size(); x++) {
	    if (list.get(x) != null) {if (list.get(x).Name.equalsIgnoreCase(match)) return x;}
	}
	return -1;
    }

    
    /**
     * 
     * @param arr
     * @param match
     * @return
     */
    public static int inChannelArray(Channel[] arr, String match) {
	if (match.length() == 0) return -1;
	    for (int x = 0; x < arr.length; x++) {
	    if (arr[x] != null) {if (arr[x].Name.equalsIgnoreCase(match)) return x;}
	}
	return -1;
    }

    private static int firstNull(Object[] o) {
	for (int x = 0; x < o.length; x++) {
	    if (o[x] == null) return x;
	}
	return -1;
    }

    class DisconnectListenerThread extends Thread {
        public boolean connected = false;

        public void run() {
            while (true) {
                try {
                    Thread.sleep(10);
                    if (connected) {
                        pongreply = false;
                        out.println("PING " + IRCServerName);
                        Thread.sleep(30000);
                        if (!pongreply) {
                            connected = false;
                            reconnect();
                        }
                    }
                } catch (Exception ex) {
                    ex.printStackTrace();
                }
            }
        }
    }
}